@@ -1,3 +1,6 @@
+# NOTE .to_s hides require from Opal
+require 'asciidoctor'.to_s unless defined? Asciidoctor
+
 # encoding: UTF-8
 module Asciidoctor
 # Extensions provide a way to participate in the parsing and converting
@@ -107,37 +110,46 @@ module Extensions
     # opts   - An optional Hash of options (default: {}):
     #          :level    - [Integer] The level to assign to this section; defaults to
     #                      one greater than the parent level (optional).
-    #          :numbered - [Boolean] A flag to force numbering, which falls back to the 
+    #          :numbered - [Boolean] A flag to force numbering, which falls back to the
     #                      state of the sectnums document attribute (optional).
     #
     # Returns a [Section] node with all properties properly initialized.
     def create_section parent, title, attrs, opts = {}
       doc = parent.document
-      doctype, level = doc.doctype, (opts[:level] || parent.level + 1)
+      book = (doctype = doc.doctype) == 'book'
+      level = opts[:level] || parent.level + 1
       if (style = attrs.delete 'style')
-        if style == 'abstract' && doctype == 'book'
+        if book && style == 'abstract'
           sectname, level = 'chapter', 1
         else
           sectname, special = style, true
           level = 1 if level == 0
         end
-      elsif doctype == 'book'
-        sectname = level == 0 ? 'part' : (level == 1 ? 'chapter' : 'section')
+      elsif book
+        sectname = level == 0 ? 'part' : (level > 1 ? 'section' : 'chapter')
       elsif doctype == 'manpage' && (title.casecmp 'synopsis') == 0
         sectname, special = 'synopsis', true
       else
         sectname = 'section'
       end
-      sect = Section.new parent, level, false
+      sect = Section.new parent, level
       sect.title, sect.sectname = title, sectname
       if special
         sect.special = true
-        sect.numbered = true if opts.fetch :numbered, (style == 'appendix')
-      elsif opts.fetch :numbered, (level > 0 && (doc.attributes.key? 'sectnums'))
-        sect.numbered = sect.special ? (parent.context == :section && parent.numbered) : true
+        if opts.fetch :numbered, (style == 'appendix')
+          sect.numbered = true
+        elsif !(opts.key? :numbered) && (doc.attr? 'sectnums', 'all')
+          sect.numbered = book && level == 1 ? :chapter : true
+        end
+      elsif level > 0
+        if opts.fetch :numbered, (doc.attr? 'sectnums')
+          sect.numbered = sect.special ? parent.numbered && true : true
+        end
+      else
+        sect.numbered = true if opts.fetch :numbered, (book && (doc.attr? 'partnums'))
       end
       unless (id = attrs.delete 'id') == false
-        sect.id = attrs['id'] = id || ((doc.attributes.key? 'sectids') ? (Section.generate_id sect.title, doc) : nil)
+        sect.id = attrs['id'] = id || ((doc.attr? 'sectids') ? (Section.generate_id sect.title, doc) : nil)
       end
       sect.update_attributes attrs
       sect
@@ -152,7 +164,7 @@ module Extensions
     # parent - The parent Block (Block, Section, or Document) of this new image block.
     # attrs  - A Hash of attributes to control how the image block is built.
     #          Use the target attribute to set the source of the image.
-    #          Use the alt attribute to specify an alternate text for the image.
+    #          Use the alt attribute to specify an alternative text for the image.
     # opts   - An optional Hash of options (default: {})
     #
     # Returns a [Block] node with all properties properly initialized.
@@ -643,13 +655,12 @@ module Extensions
     # Public: Returns the {Asciidoctor::Document} on which the extensions in this registry are being used.
     attr_reader :document
 
-    # Public: Returns the Array of {Group} classes, instances and/or Procs that have been registered.
+    # Public: Returns the Hash of {Group} classes, instances, and/or Procs that have been registered with this registry.
     attr_reader :groups
 
     def initialize groups = {}
       @groups = groups
-      @preprocessor_extensions = @tree_processor_extensions = @postprocessor_extensions = @include_processor_extensions = @docinfo_processor_extensions = nil
-      @block_extensions = @block_macro_extensions = @inline_macro_extensions = nil
+      @preprocessor_extensions = @tree_processor_extensions = @postprocessor_extensions = @include_processor_extensions = @docinfo_processor_extensions = @block_extensions = @block_macro_extensions = @inline_macro_extensions = nil
       @document = nil
     end
 
@@ -661,19 +672,21 @@ module Extensions
     # Returns the instance of this [Registry].
     def activate document
       @document = document
-      (Extensions.groups.values + @groups.values).each do |group|
-        case group
-        when ::Proc
-          case group.arity
-          when 0, -1
-            instance_exec(&group)
-          when 1
-            group.call self
+      unless (ext_groups = Extensions.groups.values + @groups.values).empty?
+        ext_groups.each do |group|
+          case group
+          when ::Proc
+            case group.arity
+            when 0, -1
+              instance_exec(&group)
+            when 1
+              group.call self
+            end
+          when ::Class
+            group.new.activate self
+          else
+            group.activate self
           end
-        when ::Class
-          group.new.activate self
-        else
-          group.activate self
         end
       end
       self
@@ -705,7 +718,7 @@ module Extensions
     #
     #   # as a method block
     #   preprocessor do
-    #     process |doc, reader|
+    #     process do |doc, reader|
     #       ...
     #     end
     #   end
@@ -757,7 +770,7 @@ module Extensions
     #
     #   # as a method block
     #   tree_processor do
-    #     process |document|
+    #     process do |document|
     #       ...
     #     end
     #   end
@@ -814,7 +827,7 @@ module Extensions
     #
     #   # as a method block
     #   postprocessor do
-    #     process |document, output|
+    #     process do |document, output|
     #       ...
     #     end
     #   end
@@ -866,7 +879,7 @@ module Extensions
     #
     #   # as a method block
     #   include_processor do
-    #     process |document, output|
+    #     process do |document, output|
     #       ...
     #     end
     #   end
@@ -918,7 +931,7 @@ module Extensions
     #
     #   # as a method block
     #   docinfo_processor do
-    #     process |doc|
+    #     process do |doc|
     #       at_location :footer
     #       'footer content'
     #     end
@@ -1007,14 +1020,14 @@ module Extensions
     #   # as a method block
     #   block do
     #     named :shout
-    #     process |parent, reader, attrs|
+    #     process do |parent, reader, attrs|
     #       ...
     #     end
     #   end
     #
     #   # as a method block with an explicit block name
     #   block :shout do
-    #     process |parent, reader, attrs|
+    #     process do |parent, reader, attrs|
     #       ...
     #     end
     #   end
@@ -1096,14 +1109,14 @@ module Extensions
     #   # as a method block
     #   block_macro do
     #     named :gist
-    #     process |parent, target, attrs|
+    #     process do |parent, target, attrs|
     #       ...
     #     end
     #   end
     #
     #   # as a method block with an explicit macro name
     #   block_macro :gist do
-    #     process |parent, target, attrs|
+    #     process do |parent, target, attrs|
     #       ...
     #     end
     #   end
@@ -1168,7 +1181,7 @@ module Extensions
     #   inline_macro ChromeInlineMacro
     #
     #   # as an InlineMacroProcessor subclass with an explicit macro name
-    #   inline_macro ChromeInineMacro, :chrome
+    #   inline_macro ChromeInlineMacro, :chrome
     #
     #   # as an instance of an InlineMacroProcessor subclass
     #   inline_macro ChromeInlineMacro.new
@@ -1180,19 +1193,19 @@ module Extensions
     #   inline_macro 'ChromeInlineMacro'
     #
     #   # as a name of an InlineMacroProcessor subclass with an explicit macro name
-    #   inline_macro 'ChromeInineMacro', :chrome
+    #   inline_macro 'ChromeInlineMacro', :chrome
     #
     #   # as a method block
     #   inline_macro do
     #     named :chrome
-    #     process |parent, target, attrs|
+    #     process do |parent, target, attrs|
     #       ...
     #     end
     #   end
     #
     #   # as a method block with an explicit macro name
     #   inline_macro :chrome do
-    #     process |parent, target, attrs|
+    #     process do |parent, target, attrs|
     #       ...
     #     end
     #   end
@@ -1477,17 +1490,36 @@ module Extensions
     # Public: Resolves the Class object for the qualified name.
     #
     # Returns Class
-    def class_for_name qualified_name
-      resolved = ::Object
-      (qualified_name.split '::').each do |name|
-        unless name.empty? || ((resolved.const_defined? name) && ::Module === (resolved = resolved.const_get name))
-          raise ::NameError, %(Could not resolve class for name: #{qualified_name})
+    if RUBY_MIN_VERSION_2
+      def class_for_name qualified_name
+        resolved = ::Object.const_get qualified_name, false
+        raise unless ::Class === resolved
+        resolved
+      rescue
+        raise ::NameError, %(Could not resolve class for name: #{qualified_name})
+      end
+    elsif RUBY_MIN_VERSION_1_9
+      def class_for_name qualified_name
+        resolved = (qualified_name.split '::').reduce ::Object do |current, name|
+          name.empty? ? current : (current.const_get name, false)
         end
+        raise unless ::Class === resolved
+        resolved
+      rescue
+        raise ::NameError, %(Could not resolve class for name: #{qualified_name})
+      end
+    else
+      def class_for_name qualified_name
+        resolved = (qualified_name.split '::').reduce ::Object do |current, name|
+          # NOTE on Ruby 1.8, const_defined? only checks for constant in current scope
+          name.empty? ? current : ((current.const_defined? name) ? (current.const_get name) : raise)
+        end
+        raise unless ::Class === resolved
+        resolved
+      rescue
+        raise ::NameError, %(Could not resolve class for name: #{qualified_name})
       end
-      raise ::NameError, %(Could not resolve class for name: #{qualified_name}) unless ::Class === resolved
-      resolved
     end
   end
-
 end
 end
