@@ -10,14 +10,17 @@ class ExtensionsInitTest < Minitest::Test
     refute doc.extensions?, 'Extensions should not be enabled by default'
 
     begin
-      # NOTE trigger extensions to autoload
-      Asciidoctor::Extensions.groups
+      # NOTE trigger extensions to autoload by registering empty group
+      Asciidoctor::Extensions.register do
+      end
     rescue; end
 
     doc = empty_document
     assert doc.extensions?, 'Extensions should be enabled after being autoloaded'
 
     self.class.remove_tests self.class
+  ensure
+    Asciidoctor::Extensions.unregister_all
   end
   self
 end.new(nil).test_autoload
@@ -200,6 +203,12 @@ end
 
 context 'Extensions' do
   context 'Register' do
+    test 'should not activate registry if no extension groups are registered' do
+      assert defined? Asciidoctor::Extensions
+      doc = empty_document
+      refute doc.extensions?, 'Extensions should not be enabled if not groups are registered'
+    end
+
     test 'should register extension group class' do
       begin
         Asciidoctor::Extensions.register :sample, SampleExtensionGroup
@@ -238,7 +247,7 @@ context 'Extensions' do
         Asciidoctor::Extensions.register :sample, SampleExtensionGroup.new
         refute_nil Asciidoctor::Extensions.groups
         assert_equal 1, Asciidoctor::Extensions.groups.size
-        assert Asciidoctor::Extensions.groups[:sample].is_a? SampleExtensionGroup
+        assert_kind_of SampleExtensionGroup, Asciidoctor::Extensions.groups[:sample]
       ensure
         Asciidoctor::Extensions.unregister_all
       end
@@ -250,7 +259,7 @@ context 'Extensions' do
         end
         refute_nil Asciidoctor::Extensions.groups
         assert_equal 1, Asciidoctor::Extensions.groups.size
-        assert Asciidoctor::Extensions.groups[:sample].is_a? Proc
+        assert_kind_of Proc, Asciidoctor::Extensions.groups[:sample]
       ensure
         Asciidoctor::Extensions.unregister_all
       end
@@ -331,6 +340,24 @@ context 'Extensions' do
       end
     end
 
+    test 'should raise exception if constant name is invalid' do
+      begin
+        Asciidoctor::Extensions.class_for_name 'foobar'
+        flunk 'Expecting RuntimeError to be raised'
+      rescue NameError => e
+        assert_equal 'Could not resolve class for name: foobar', e.message
+      end
+    end
+
+    test 'should raise exception if class not found in scope' do
+      begin
+        Asciidoctor::Extensions.class_for_name 'Asciidoctor::Extensions::String'
+        flunk 'Expecting RuntimeError to be raised'
+      rescue NameError => e
+        assert_equal 'Could not resolve class for name: Asciidoctor::Extensions::String', e.message
+      end
+    end
+
     test 'should raise exception if name resolves to module' do
       begin
         Asciidoctor::Extensions.class_for_name 'Asciidoctor::Extensions'
@@ -352,6 +379,29 @@ context 'Extensions' do
       assert_equal Asciidoctor::Document, clazz
     end
 
+    test 'should not resolve class if not in scope' do
+      begin
+        Asciidoctor::Extensions.resolve_class 'Asciidoctor::Extensions::String'
+        flunk 'Expecting RuntimeError to be raised'
+      rescue NameError => e
+        assert_equal 'Could not resolve class for name: Asciidoctor::Extensions::String', e.message
+      end
+    end
+
+    test 'should raise NameError if extension class cannot be resolved from string' do
+      begin
+        Asciidoctor::Extensions.register do
+          block 'foobar'
+        end
+        empty_document
+        flunk 'Expecting RuntimeError to be raised'
+      rescue NameError => e
+        assert_equal 'Could not resolve class for name: foobar', e.message
+      ensure
+        Asciidoctor::Extensions.unregister_all
+      end
+    end
+
     test 'should allow standalone registry to be created but not registered' do
       registry = Asciidoctor::Extensions.create 'sample' do
         block do
@@ -407,7 +457,7 @@ context 'Extensions' do
         SampleExtensionGroup.register
         doc = Asciidoctor::Document.new
         assert doc.extensions?
-        assert doc.extensions.is_a? Asciidoctor::Extensions::Registry
+        assert_kind_of Asciidoctor::Extensions::Registry, doc.extensions
       ensure
         Asciidoctor::Extensions.unregister_all
       end
@@ -423,9 +473,9 @@ context 'Extensions' do
       assert registry.preprocessors?
       extensions = registry.preprocessors
       assert_equal 1, extensions.size
-      assert extensions.first.is_a? Asciidoctor::Extensions::ProcessorExtension
-      assert extensions.first.instance.is_a? SamplePreprocessor
-      assert extensions.first.process_method.is_a? ::Method
+      assert_kind_of Asciidoctor::Extensions::ProcessorExtension, extensions.first
+      assert_kind_of SamplePreprocessor, extensions.first.instance
+      assert_kind_of Method, extensions.first.process_method
     end
 
     test 'should instantiate include processors' do
@@ -435,9 +485,9 @@ context 'Extensions' do
       assert registry.include_processors?
       extensions = registry.include_processors
       assert_equal 1, extensions.size
-      assert extensions.first.is_a? Asciidoctor::Extensions::ProcessorExtension
-      assert extensions.first.instance.is_a? SampleIncludeProcessor
-      assert extensions.first.process_method.is_a? ::Method
+      assert_kind_of Asciidoctor::Extensions::ProcessorExtension, extensions.first
+      assert_kind_of SampleIncludeProcessor, extensions.first.instance
+      assert_kind_of Method, extensions.first.process_method
     end
 
     test 'should instantiate docinfo processors' do
@@ -448,9 +498,9 @@ context 'Extensions' do
       assert registry.docinfo_processors?(:head)
       extensions = registry.docinfo_processors
       assert_equal 1, extensions.size
-      assert extensions.first.is_a? Asciidoctor::Extensions::ProcessorExtension
-      assert extensions.first.instance.is_a? SampleDocinfoProcessor
-      assert extensions.first.process_method.is_a? ::Method
+      assert_kind_of Asciidoctor::Extensions::ProcessorExtension, extensions.first
+      assert_kind_of SampleDocinfoProcessor, extensions.first.instance
+      assert_kind_of Method, extensions.first.process_method
     end
 
     # NOTE intentionally using the legacy names
@@ -461,9 +511,9 @@ context 'Extensions' do
       assert registry.treeprocessors?
       extensions = registry.treeprocessors
       assert_equal 1, extensions.size
-      assert extensions.first.is_a? Asciidoctor::Extensions::ProcessorExtension
-      assert extensions.first.instance.is_a? SampleTreeprocessor
-      assert extensions.first.process_method.is_a? ::Method
+      assert_kind_of Asciidoctor::Extensions::ProcessorExtension, extensions.first
+      assert_kind_of SampleTreeprocessor, extensions.first.instance
+      assert_kind_of Method, extensions.first.process_method
     end
 
     test 'should instantiate postprocessors' do
@@ -473,9 +523,9 @@ context 'Extensions' do
       assert registry.postprocessors?
       extensions = registry.postprocessors
       assert_equal 1, extensions.size
-      assert extensions.first.is_a? Asciidoctor::Extensions::ProcessorExtension
-      assert extensions.first.instance.is_a? SamplePostprocessor
-      assert extensions.first.process_method.is_a? ::Method
+      assert_kind_of Asciidoctor::Extensions::ProcessorExtension, extensions.first
+      assert_kind_of SamplePostprocessor, extensions.first.instance
+      assert_kind_of Method, extensions.first.process_method
     end
 
     test 'should instantiate block processor' do
@@ -485,9 +535,9 @@ context 'Extensions' do
       assert registry.blocks?
       assert registry.registered_for_block? :sample, :paragraph
       extension = registry.find_block_extension :sample
-      assert extension.is_a? Asciidoctor::Extensions::ProcessorExtension
-      assert extension.instance.is_a? SampleBlock
-      assert extension.process_method.is_a? ::Method
+      assert_kind_of Asciidoctor::Extensions::ProcessorExtension, extension
+      assert_kind_of SampleBlock, extension.instance
+      assert_kind_of Method, extension.process_method
     end
 
     test 'should not match block processor for unsupported context' do
@@ -504,9 +554,9 @@ context 'Extensions' do
       assert registry.block_macros?
       assert registry.registered_for_block_macro? 'sample'
       extension = registry.find_block_macro_extension 'sample'
-      assert extension.is_a? Asciidoctor::Extensions::ProcessorExtension
-      assert extension.instance.is_a? SampleBlockMacro
-      assert extension.process_method.is_a? ::Method
+      assert_kind_of Asciidoctor::Extensions::ProcessorExtension, extension
+      assert_kind_of SampleBlockMacro, extension.instance
+      assert_kind_of Method, extension.process_method
     end
 
     test 'should instantiate inline macro processor' do
@@ -516,9 +566,9 @@ context 'Extensions' do
       assert registry.inline_macros?
       assert registry.registered_for_inline_macro? 'sample'
       extension = registry.find_inline_macro_extension 'sample'
-      assert extension.is_a? Asciidoctor::Extensions::ProcessorExtension
-      assert extension.instance.is_a? SampleInlineMacro
-      assert extension.process_method.is_a? ::Method
+      assert_kind_of Asciidoctor::Extensions::ProcessorExtension, extension
+      assert_kind_of SampleInlineMacro, extension.instance
+      assert_kind_of Method, extension.process_method
     end
 
     test 'should allow processors to be registered by a string name' do
@@ -528,12 +578,12 @@ context 'Extensions' do
       assert registry.preprocessors?
       extensions = registry.preprocessors
       assert_equal 1, extensions.size
-      assert extensions.first.is_a? Asciidoctor::Extensions::ProcessorExtension
+      assert_kind_of Asciidoctor::Extensions::ProcessorExtension, extensions.first
     end
   end
 
   context 'Integration' do
-    test 'can provide extension registry as option' do
+    test 'can provide extension registry as an option' do
       registry = Asciidoctor::Extensions.create do
         tree_processor SampleTreeProcessor
       end
@@ -546,6 +596,19 @@ context 'Extensions' do
       assert_equal 0, Asciidoctor::Extensions.groups.size
     end
 
+    # NOTE I'm not convinced we want to continue to support this use case
+    test 'can provide extension registry created without any groups as option' do
+      registry = Asciidoctor::Extensions.create
+      registry.tree_processor SampleTreeProcessor
+
+      doc = document_from_string %(= Document Title\n\ncontent), :extension_registry => registry
+      refute_nil doc.extensions
+      assert_equal 0, doc.extensions.groups.size
+      assert doc.extensions.tree_processors?
+      assert_equal 1, doc.extensions.tree_processors.size
+      assert_equal 0, Asciidoctor::Extensions.groups.size
+    end
+
     test 'can provide extensions proc as option' do
       doc = document_from_string %(= Document Title\n\ncontent), :extensions => proc {
         tree_processor SampleTreeProcessor
@@ -597,9 +660,9 @@ after
 
         result = render_string input, :safe => :server
         assert_css '.paragraph > p', result, 3
-        assert result.include?('before')
-        assert result.include?('Lorem ipsum')
-        assert result.include?('after')
+        assert_includes result, 'before'
+        assert_includes result, 'Lorem ipsum'
+        assert_includes result, 'after'
       ensure
         Asciidoctor::Extensions.unregister_all
       end
@@ -614,19 +677,21 @@ include::include-file.asciidoc[]
 last line
       EOS
 
-      # Safe Mode is not required here
-      document = empty_document :base_dir => File.expand_path(File.dirname(__FILE__))
-      document.extensions.include_processor do
-        handles? do |target|
-          target == 'include-file.asciidoc'
-        end
+      registry = Asciidoctor::Extensions.create do
+        include_processor do
+          handles? do |target|
+            target == 'include-file.asciidoc'
+          end
 
-        process do |doc, reader, target, attributes|
-          # demonstrate that push_include normalizes endlines
-          content = ["include target:: #{target}\n", "\n", "middle line\n"]
-          reader.push_include content, target, target, 1, attributes
+          process do |doc, reader, target, attributes|
+            # demonstrate that push_include normalizes endlines
+            content = ["include target:: #{target}\n", "\n", "middle line\n"]
+            reader.push_include content, target, target, 1, attributes
+          end
         end
       end
+      # Safe Mode is not required here
+      document = empty_document :base_dir => testdir, :extension_registry => registry
       reader = Asciidoctor::PreprocessorReader.new document, input, nil, :normalize => true
       lines = []
       lines << reader.read_line
@@ -662,6 +727,25 @@ content
       end
     end
 
+    test 'should set source_location on document before invoking tree processors' do
+      begin
+        Asciidoctor::Extensions.register do
+          tree_processor do
+            process do |doc|
+              para = create_paragraph doc.blocks.last.parent, %(file: #{doc.file}, lineno: #{doc.lineno}), {}
+              doc << para
+            end
+          end
+        end
+
+        sample_doc = fixture_path 'sample.asciidoc'
+        doc = Asciidoctor.load_file sample_doc, :sourcemap => true
+        assert_includes doc.convert, 'file: sample.asciidoc, lineno: 1'
+      ensure
+        Asciidoctor::Extensions.unregister_all
+      end
+    end
+
     test 'should allow tree processor to replace tree' do
       input = <<-EOS
 = Original Document
@@ -751,6 +835,28 @@ Hi there!
       end
     end
 
+    test 'should invoke processor for custom block in an AsciiDoc table cell' do
+      input = <<-EOS
+|===
+a|
+[yell]
+Hi there!
+|===
+      EOS
+
+      begin
+        Asciidoctor::Extensions.register do
+          block UppercaseBlock
+        end
+
+        output = render_embedded_string input
+        assert_xpath '/table//p', output, 1
+        assert_xpath '/table//p[text()="HI THERE!"]', output, 1
+      ensure
+        Asciidoctor::Extensions.unregister_all
+      end
+    end
+
     test 'should pass cloaked context in attributes passed to process method of custom block' do
       input = <<-EOS
 [custom]
@@ -789,7 +895,30 @@ snippet::12345[mode=edit]
         end
 
         output = render_embedded_string input
-        assert output.include?('<script src="http://example.com/12345.js?_mode=edit"></script>')
+        assert_includes output, '<script src="http://example.com/12345.js?_mode=edit"></script>'
+      ensure
+        Asciidoctor::Extensions.unregister_all
+      end
+    end
+
+    test 'should invoke processor for custom block macro in an AsciiDoc table cell' do
+      input = <<-EOS
+|===
+a|message::hi[]
+|===
+      EOS
+
+      begin
+        Asciidoctor::Extensions.register do
+          block_macro :message do
+            process do |parent, target, attrs|
+              create_paragraph parent, target.upcase, {}
+            end
+          end
+        end
+
+        output = render_embedded_string input
+        assert_xpath '/table//p[text()="HI"]', output, 1
       ensure
         Asciidoctor::Extensions.unregister_all
       end
@@ -828,10 +957,10 @@ custom_toc::[]
         end
 
         output = render_embedded_string 'Room temperature is deg:25[C,precision=0].', :attributes => { 'temperature-unit' => 'F' }
-        assert output.include?('Room temperature is 25 &#176;C.')
+        assert_includes output, 'Room temperature is 25 &#176;C.'
 
         output = render_embedded_string 'Normal body temperature is deg:37[].', :attributes => { 'temperature-unit' => 'F' }
-        assert output.include?('Normal body temperature is 98.6 &#176;F.')
+        assert_includes output, 'Normal body temperature is 98.6 &#176;F.'
       ensure
         Asciidoctor::Extensions.unregister_all
       end
@@ -851,7 +980,7 @@ custom_toc::[]
         end
 
         output = render_embedded_string 'label:[Checkbox]'
-        assert output.include?('<label>Checkbox</label>')
+        assert_includes output, '<label>Checkbox</label>'
       ensure
         Asciidoctor::Extensions.unregister_all
       end
@@ -952,7 +1081,7 @@ target="target", attributes=[]
         end
 
         output = render_embedded_string 'mention:mojavelinux[Dan]'
-        assert output.include?('<a href="https://github.com/mojavelinux">Dan</a>')
+        assert_includes output, '<a href="https://github.com/mojavelinux">Dan</a>'
       ensure
         Asciidoctor::Extensions.unregister_all
       end
@@ -1094,6 +1223,7 @@ content
             process do |parent, target, attrs|
               opts = (level = attrs.delete 'level') ? { :level => level.to_i } : {}
               attrs['id'] = false if attrs['id'] == 'false'
+              parent = parent.parent if parent.context == :preamble
               sect = create_section parent, 'Section Title', attrs, opts
               nil
             end
@@ -1111,20 +1241,26 @@ sect::[%s]
         {
           ''                       => ['chapter',  1, false, true, '_section_title'],
           'level=0'                => ['part',     0, false, false, '_section_title'],
+          'level=0,alt'            => ['part',     0, false, true, '_section_title', { 'partnums' => '' }],
           'level=0,style=appendix' => ['appendix', 1, true,  true, '_section_title'],
           'style=appendix'         => ['appendix', 1, true,  true, '_section_title'],
           'style=glossary'         => ['glossary', 1, true,  false, '_section_title'],
+          'style=glossary,alt'     => ['glossary', 1, true,  :chapter, '_section_title', { 'sectnums' => 'all' }],
           'style=abstract'         => ['chapter',  1, false, true, '_section_title'],
           'id=section-title'       => ['chapter',  1, false, true, 'section-title'],
           'id=false'               => ['chapter',  1, false, true, nil]
-        }.each do |attrlist, (expect_sectname, expect_level, expect_special, expect_numbered, expect_id)|
+        }.each do |attrlist, (expect_sectname, expect_level, expect_special, expect_numbered, expect_id, extra_attrs)|
           input = input_tpl % attrlist
-          document_from_string input, :safe => :server
+          document_from_string input, :safe => :server, :attributes => extra_attrs
           assert_equal expect_sectname, sect.sectname
           assert_equal expect_level, sect.level
           assert_equal expect_special, sect.special
           assert_equal expect_numbered, sect.numbered
-          assert_equal expect_id, sect.id
+          if expect_id
+            assert_equal expect_id, sect.id
+          else
+            assert_nil sect.id
+          end
         end
       ensure
         Asciidoctor::Extensions.unregister_all
@@ -1150,7 +1286,6 @@ sample content
       end
     end
 
-
     test 'should add multiple docinfo to document' do
       input = <<-EOS
 = Document Title
