@@ -33,8 +33,8 @@ context 'Invoker' do
   end
 
   test 'should set implicit doc info attributes' do
-    sample_filepath = File.expand_path(File.join(File.dirname(__FILE__), 'fixtures', 'sample.asciidoc'))
-    sample_filedir = File.expand_path(File.join(File.dirname(__FILE__), 'fixtures'))
+    sample_filepath = fixture_path 'sample.asciidoc'
+    sample_filedir = fixturedir
     invoker = invoke_cli_to_buffer %w(-o /dev/null), sample_filepath
     doc = invoker.document
     assert_equal 'sample', doc.attr('docname')
@@ -48,7 +48,7 @@ context 'Invoker' do
   end
 
   test 'should allow docdate and doctime to be overridden' do
-    sample_filepath = File.expand_path(File.join(File.dirname(__FILE__), 'fixtures', 'sample.asciidoc'))
+    sample_filepath = fixture_path 'sample.asciidoc'
     invoker = invoke_cli_to_buffer %w(-o /dev/null -a docdate=2015-01-01 -a doctime=10:00:00-07:00), sample_filepath
     doc = invoker.document
     assert doc.attr?('docdate', '2015-01-01')
@@ -86,7 +86,7 @@ context 'Invoker' do
   end
 
   test 'should accept document from stdin and write to output file' do
-    sample_outpath = File.expand_path(File.join(File.dirname(__FILE__), 'fixtures', 'sample-output.html'))
+    sample_outpath = fixture_path 'sample-output.html'
     begin
       invoker = invoke_cli(%W(-s -o #{sample_outpath}), '-') { 'content' }
       doc = invoker.document
@@ -111,13 +111,13 @@ context 'Invoker' do
   end
 
   test 'should fail if input file matches specified output file' do
-    sample_outpath = File.expand_path(File.join(File.dirname(__FILE__), 'fixtures', 'sample.asciidoc'))
+    sample_outpath = fixture_path 'sample.asciidoc'
     invoker = invoke_cli_to_buffer %W(-o #{sample_outpath}), 'sample.asciidoc'
     assert_match(/input file and output file cannot be the same/, invoker.read_error)
   end
 
   test 'should accept input from named pipe and output to stdout' do
-    sample_inpath = File.expand_path(File.join(File.dirname(__FILE__), 'fixtures', 'sample-pipe.adoc'))
+    sample_inpath = fixture_path 'sample-pipe.adoc'
     begin
       %x(mkfifo #{sample_inpath})
       write_thread = Thread.new do
@@ -133,7 +133,7 @@ context 'Invoker' do
   end if RUBY_MIN_VERSION_1_9 && !windows?
 
   test 'should allow docdir to be specified when input is a string' do
-    expected_docdir = File.expand_path(File.join(File.dirname(__FILE__), 'fixtures'))
+    expected_docdir = fixturedir
     invoker = invoke_cli_to_buffer(%w(-s --base-dir test/fixtures -o /dev/null), '-') { 'content' }
     doc = invoker.document
     assert_equal expected_docdir, doc.attr('docdir')
@@ -141,7 +141,7 @@ context 'Invoker' do
   end
 
   test 'should display version and exit' do
-    expected = %(Asciidoctor #{Asciidoctor::VERSION} [http://asciidoctor.org]\nRuntime Environment (#{RUBY_DESCRIPTION}))
+    expected = %(Asciidoctor #{Asciidoctor::VERSION} [https://asciidoctor.org]\nRuntime Environment (#{RUBY_DESCRIPTION}))
     ['--version', '-V'].each do |switch|
       actual = nil
       redirect_streams do |out, err|
@@ -179,6 +179,18 @@ context 'Invoker' do
     assert_equal '', warnings
   end
 
+  test 'should return non-zero exit code if failure level is reached' do
+    input = <<-EOS
+2. second
+3. third
+    EOS
+    exit_code, messages = redirect_streams do |_, err|
+      [invoke_cli(%w(-q --failure-level=WARN -o /dev/null), '-') { input }.code, err.string]
+    end
+    assert_equal 1, exit_code
+    assert messages.empty?
+  end
+
   test 'should report usage if no input file given' do
     redirect_streams do |out, err|
       invoke_cli [], nil
@@ -189,7 +201,7 @@ context 'Invoker' do
   test 'should report error if input file does not exist' do
     redirect_streams do |out, err|
       invoker = invoke_cli [], 'missing_file.asciidoc'
-      assert_match(/input file .* missing or cannot be read/, err.string)
+      assert_match(/input file .* is missing/, err.string)
       assert_equal 1, invoker.code
     end
   end
@@ -197,13 +209,13 @@ context 'Invoker' do
   test 'should treat extra arguments as files' do
     redirect_streams do |out, err|
       invoker = invoke_cli %w(-o /dev/null extra arguments sample.asciidoc), nil
-      assert_match(/input file .* missing or cannot be read/, err.string)
+      assert_match(/input file .* is missing/, err.string)
       assert_equal 1, invoker.code
     end
   end
 
   test 'should output to file name based on input file name' do
-    sample_outpath = File.expand_path(File.join(File.dirname(__FILE__), 'fixtures', 'sample.html'))
+    sample_outpath = fixture_path 'sample.html'
     begin
       invoker = invoke_cli
       doc = invoker.document
@@ -222,8 +234,8 @@ context 'Invoker' do
   end
 
   test 'should output to file in destination directory if set' do
-    destination_path = File.expand_path(File.join(File.dirname(__FILE__), 'test_output'))
-    sample_outpath = File.join(destination_path, 'sample.html')
+    destination_path = File.join testdir, 'test_output'
+    sample_outpath = File.join destination_path, 'sample.html'
     begin
       FileUtils.mkdir_p(destination_path)
       # QUESTION should -D be relative to working directory or source directory?
@@ -238,8 +250,25 @@ context 'Invoker' do
     end
   end
 
+  test 'should preserve directory structure in destination directory if source directory is set' do
+    sample_inpath = 'subdir/index.adoc'
+    destination_path = 'test_output'
+    destination_subdir_path = File.join destination_path, 'subdir'
+    sample_outpath = File.join destination_subdir_path, 'index.html'
+    begin
+      FileUtils.mkdir_p(destination_path)
+      invoke_cli %W(-D #{destination_path} -R test/fixtures), sample_inpath
+      assert File.directory?(destination_subdir_path)
+      assert File.exist?(sample_outpath)
+    ensure
+      FileUtils.rm_f(sample_outpath)
+      FileUtils.rmdir(destination_subdir_path)
+      FileUtils.rmdir(destination_path)
+    end
+  end
+
   test 'should output to file specified' do
-    sample_outpath = File.expand_path(File.join(File.dirname(__FILE__), 'fixtures', 'sample-output.html'))
+    sample_outpath = fixture_path 'sample-output.html'
     begin
       invoker = invoke_cli %W(-o #{sample_outpath})
       doc = invoker.document
@@ -251,9 +280,9 @@ context 'Invoker' do
   end
 
   test 'should copy default stylesheet to target directory if linkcss is specified' do
-    sample_outpath = File.expand_path(File.join(File.dirname(__FILE__), 'fixtures', 'sample-output.html'))
-    asciidoctor_stylesheet = File.expand_path(File.join(File.dirname(__FILE__), 'fixtures', 'asciidoctor.css'))
-    coderay_stylesheet = File.expand_path(File.join(File.dirname(__FILE__), 'fixtures', 'coderay-asciidoctor.css'))
+    sample_outpath = fixture_path 'sample-output.html'
+    asciidoctor_stylesheet = fixture_path 'asciidoctor.css'
+    coderay_stylesheet = fixture_path 'coderay-asciidoctor.css'
     begin
       invoker = invoke_cli %W(-o #{sample_outpath} -a linkcss -a source-highlighter=coderay)
       invoker.document
@@ -268,8 +297,8 @@ context 'Invoker' do
   end
 
   test 'should not copy default stylesheet to target directory if linkcss is set and copycss is unset' do
-    sample_outpath = File.expand_path(File.join(File.dirname(__FILE__), 'fixtures', 'sample-output.html'))
-    default_stylesheet = File.expand_path(File.join(File.dirname(__FILE__), 'fixtures', 'asciidoctor.css'))
+    sample_outpath = fixture_path 'sample-output.html'
+    default_stylesheet = fixture_path 'asciidoctor.css'
     begin
       invoker = invoke_cli %W(-o #{sample_outpath} -a linkcss -a copycss!)
       invoker.document
@@ -282,7 +311,7 @@ context 'Invoker' do
   end
 
   test 'should copy custom stylesheet to target directory if stylesheet and linkcss is specified' do
-    destdir = File.expand_path(File.join(File.dirname(__FILE__), 'fixtures', 'output'))
+    destdir = fixture_path 'output'
     sample_outpath = File.join destdir, 'sample-output.html'
     stylesdir = File.join destdir, 'styles'
     custom_stylesheet = File.join stylesdir, 'custom.css'
@@ -300,7 +329,7 @@ context 'Invoker' do
   end
 
   test 'should not copy custom stylesheet to target directory if stylesheet and linkcss are set and copycss is unset' do
-    destdir = File.expand_path(File.join(File.dirname(__FILE__), 'fixtures', 'output'))
+    destdir = fixture_path 'output'
     sample_outpath = File.join destdir, 'sample-output.html'
     stylesdir = File.join destdir, 'styles'
     custom_stylesheet = File.join stylesdir, 'custom.css'
@@ -318,7 +347,7 @@ context 'Invoker' do
   end
 
   test 'should not copy custom stylesheet to target directory if stylesdir is a URI' do
-    destdir = File.expand_path(File.join(File.dirname(__FILE__), 'fixtures', 'output'))
+    destdir = fixture_path 'output'
     sample_outpath = File.join destdir, 'sample-output.html'
     stylesdir = File.join destdir, 'http:'
     begin
@@ -334,8 +363,8 @@ context 'Invoker' do
   end
 
   test 'should render all passed files' do
-    basic_outpath = File.expand_path(File.join(File.dirname(__FILE__), 'fixtures', 'basic.html'))
-    sample_outpath = File.expand_path(File.join(File.dirname(__FILE__), 'fixtures', 'sample.html'))
+    basic_outpath = fixture_path 'basic.html'
+    sample_outpath = fixture_path 'sample.html'
     begin
       invoke_cli_with_filenames [], %w(basic.asciidoc sample.asciidoc)
       assert File.exist?(basic_outpath)
@@ -347,9 +376,9 @@ context 'Invoker' do
   end
 
   test 'options should not be modified when processing multiple files' do
-    destination_path = File.expand_path(File.join(File.dirname(__FILE__), 'test_output'))
-    basic_outpath = File.join(destination_path, 'basic.htm')
-    sample_outpath = File.join(destination_path, 'sample.htm')
+    destination_path = File.join testdir, 'test_output'
+    basic_outpath = File.join destination_path, 'basic.htm'
+    sample_outpath = File.join destination_path, 'sample.htm'
     begin
       invoke_cli_with_filenames %w(-D test/test_output -a outfilesuffix=.htm), %w(basic.asciidoc sample.asciidoc)
       assert File.exist?(basic_outpath)
@@ -362,7 +391,7 @@ context 'Invoker' do
   end
 
   test 'should render all files that matches a glob expression' do
-    basic_outpath = File.expand_path(File.join(File.dirname(__FILE__), 'fixtures', 'basic.html'))
+    basic_outpath = fixture_path 'basic.html'
     begin
       invoke_cli_to_buffer [], "ba*.asciidoc"
       assert File.exist?(basic_outpath)
@@ -372,8 +401,8 @@ context 'Invoker' do
   end
 
   test 'should render all files that matches an absolute path glob expression' do
-    basic_outpath = File.expand_path(File.join(File.dirname(__FILE__), 'fixtures', 'basic.html'))
-    glob = File.join(File.expand_path(File.dirname(__FILE__)), 'fixtures', 'ba*.asciidoc')
+    basic_outpath = fixture_path 'basic.html'
+    glob = fixture_path 'ba*.asciidoc'
     # test Windows using backslash-style pathname
     if File::ALT_SEPARATOR == '\\'
       glob = glob.tr '/', '\\'
@@ -394,6 +423,38 @@ context 'Invoker' do
     assert_xpath '/*[@id="preamble"]', output, 1
   end
 
+  test 'should write page for each alternate manname' do
+    outdir = fixturedir
+    outfile_1 = File.join outdir, 'eve.1'
+    outfile_2 = File.join outdir, 'islifeform.1'
+    input = <<-EOS
+= eve(1)
+Andrew Stanton
+v1.0.0
+:doctype: manpage
+:manmanual: EVE
+:mansource: EVE
+
+== NAME
+
+eve, islifeform - analyzes an image to determine if it's a picture of a life form
+
+== SYNOPSIS
+
+*eve* ['OPTION']... 'FILE'...
+    EOS
+
+    begin
+      invoke_cli(%W(-b manpage -o #{outfile_1}), '-') { input }
+      assert File.exist?(outfile_1)
+      assert File.exist?(outfile_2)
+      assert_equal '.so eve.1', (IO.read outfile_2).chomp
+    ensure
+      FileUtils.rm_f outfile_1
+      FileUtils.rm_f outfile_2
+    end
+  end
+
   test 'should output a trailing endline to stdout' do
     invoker = nil
     output = nil
@@ -467,18 +528,18 @@ context 'Invoker' do
   end
 
   test 'should locate custom templates based on template dir, template engine and backend' do
-    custom_backend_root = File.expand_path(File.join(File.dirname(__FILE__), 'fixtures', 'custom-backends'))
+    custom_backend_root = fixture_path 'custom-backends'
     invoker = invoke_cli_to_buffer %W(-E haml -T #{custom_backend_root} -o -)
     doc = invoker.document
-    assert doc.converter.is_a? Asciidoctor::Converter::CompositeConverter
+    assert_kind_of Asciidoctor::Converter::CompositeConverter, doc.converter
     selected = doc.converter.find_converter 'paragraph'
-    assert selected.is_a? Asciidoctor::Converter::TemplateConverter
-    assert selected.templates['paragraph'].is_a? Tilt::HamlTemplate
+    assert_kind_of Asciidoctor::Converter::TemplateConverter, selected
+    assert_kind_of Tilt::HamlTemplate, selected.templates['paragraph']
   end
 
   test 'should load custom templates from multiple template directories' do
-    custom_backend_1 = File.expand_path(File.join(File.dirname(__FILE__), 'fixtures', 'custom-backends/haml/html5'))
-    custom_backend_2 = File.expand_path(File.join(File.dirname(__FILE__), 'fixtures', 'custom-backends/haml/html5-tweaks'))
+    custom_backend_1 = fixture_path 'custom-backends/haml/html5'
+    custom_backend_2 = fixture_path 'custom-backends/haml/html5-tweaks'
     invoker = invoke_cli_to_buffer %W(-T #{custom_backend_1} -T #{custom_backend_2} -o - -s)
     output = invoker.read_output
     assert_css '.paragraph', output, 0
@@ -568,7 +629,7 @@ context 'Invoker' do
   end
 
   test 'should force default external encoding to UTF-8' do
-    executable = File.expand_path(File.join(File.dirname(__FILE__), '..', 'bin', 'asciidoctor'))
+    executable = File.join ASCIIDOCTOR_PROJECT_DIR, 'bin', 'asciidoctor'
     input_path = fixture_path 'encoding.asciidoc'
     old_lang = ENV['LANG']
     ENV['LANG'] = 'US-ASCII'
@@ -586,7 +647,7 @@ context 'Invoker' do
       refute_empty stdout_lines
       stdout_lines.each {|l| l.force_encoding Encoding::UTF_8 } if Asciidoctor::FORCE_ENCODING
       stdout_str = stdout_lines.join
-      assert stdout_str.include?('Codierungen sind verrückt auf älteren Versionen von Ruby')
+      assert_includes stdout_str, 'Codierungen sind verrückt auf älteren Versionen von Ruby'
     ensure
       ENV['LANG'] = old_lang
     end
@@ -611,7 +672,7 @@ context 'Invoker' do
     old_source_date_epoch = ENV.delete 'SOURCE_DATE_EPOCH'
     begin
       ENV['SOURCE_DATE_EPOCH'] = '1234123412'
-      sample_filepath = File.expand_path(File.join(File.dirname(__FILE__), 'fixtures', 'sample.asciidoc'))
+      sample_filepath = fixture_path 'sample.asciidoc'
       invoker = invoke_cli_to_buffer %w(-o /dev/null), sample_filepath
       doc = invoker.document
       assert_equal '2009-02-08', (doc.attr 'docdate')
@@ -633,7 +694,7 @@ context 'Invoker' do
     old_source_date_epoch = ENV.delete 'SOURCE_DATE_EPOCH'
     begin
       ENV['SOURCE_DATE_EPOCH'] = 'aaaaaaaa'
-      sample_filepath = File.expand_path(File.join(File.dirname(__FILE__), 'fixtures', 'sample.asciidoc'))
+      sample_filepath = fixture_path 'sample.asciidoc'
       assert_equal 1, (invoke_cli_to_buffer %w(-o /dev/null), sample_filepath).code
     ensure
       if old_source_date_epoch
